/* AI.c - Pretend that the computer can think. */
#include <stdio.h>
#include <SDL.h>
#include "blockit.h"
#include "game.h"
#include "rungame.h"
#include "ai.h"

#define AI_NASTINESS 3

/* Counts the number of squares of territory that PLAYER would
 * gain if he/she/it moved to (x,y). Returns -1 if it is not
 * a legit. move, or if it touches enemy territory.
 */
static int ai_count_terr(int player, int x, int y)
{
  int i, j;
  int c = 0;

  if (!g_can_move(player, x, y)) return -1;
  for (i=x-1; i<=x+1; i++)
    {
      if (i<0) continue;
      if (i>=G_WIDTH) continue;
      for (j=y-1; j<=y+1; j++)
	{
	  if (j<0) continue;
	  if (j>=G_HEIGHT) continue;
	  if ((g_grid[i][j] & 3) == GP_NEUTRAL) c++;
	}
    }
  return c;
}

/* Checks to see if a move would impinge upon
 * an opponent's territory.
 */
static int ai_would_conflict(int player, int x, int y)
{
  int i, j;

  if (!g_can_move(player, x, y)) return -1;
  for (i=x-1; i<=x+1; i++)
    {
      if (i<0) continue;
      if (i>=G_WIDTH) continue;
      for (j=y-1; j<=y+1; j++)
	{
	  if (j<0) continue;
	  if (j>=G_HEIGHT) continue;
	  if ((g_grid[i][j] & 3) == (3-player)) return 1; /* Opponent territory */
	}
    }
  return 0;
}

/* Counts the number of enemy stones the PLAYER would gain if
 * he made move  (x,y).
 */
int ai_count_gain(int player, int x, int y)
{
  int i,j,c=0;

  if (!g_can_move(player, x, y)) return -1;
  for (i=x-1; i<=x+1; i++)
    {
      if (i<0) continue;
      if (i>=G_WIDTH) continue;
      for (j=y-1; j<=y+1; j++)
	{
	  if (j<0) continue;
	  if (j>=G_HEIGHT) continue;
	  if ((g_grid[i][j] & GT_STONE))
	    {
	      if ((g_grid[i][j] & 3) == (3-player)) /* i.e., is owned by opponent */
		{
		  c++;
		}
	    }
	}
    }
  return c;
}

/* EASY STRATEGY
 *
 *  1. Make the move that would gain the
 *     most territory.
 *     (i.e., stay away from the edges, and from opponent's territory to a lesser extent.)
 *  2. If there are any shared squares,
 *     move there.
 *     (i.e., retaliate if necessary.)
 *
 *  Point #1 is used only if #2 fails.
 *  If there are multiple squares satisfying
 *  the criterion, one is chosen at random.
 *
 */
static void ai_easy_move(int player, int *x, int *y)
{
  int mvx=-1, mvy=-1;
  int nshared = 0;
  int maxterr = -2, nmax = 0;
  int nc, N;
  int i, j;

  for (i=0; i<G_WIDTH; i++)
    {
      for (j=0; j<G_HEIGHT; j++)
	{
	  if ((g_grid[i][j] & 3) == GP_SHARED) /* &3 to eliminate GT_STONE */
	    {
	      nshared++;
	    }
	}
    }
  if (nshared == 0) /* Point#2 couldn't be used */
    {
      for (i=0; i<G_WIDTH; i++)
	{
	  for (j=0; j<G_HEIGHT; j++)
	    {
	      nc = ai_count_terr(player, i, j);
	      if (nc > maxterr) maxterr = nc;
	    }
	} /* Determine maximum territory gain */
      for (i=0; i<G_WIDTH; i++)
	{
	  for (j=0; j<G_HEIGHT; j++)
	    {
	      nc = ai_count_terr(player, i, j);
	      if (nc == maxterr) nmax ++;
	    }
	} /* Count number of squares for randomness */
      if (nmax == 1) nc = 1;
      else nc = (rand() % (nmax))+1;
      N = 0;

      if (maxterr < 0) /* There is a problem */
	SDLhelp_Problem(bl_screen, "Argh! Bug in EASY-level AI!");

      for (i=0; i<G_WIDTH; i++)
	{
	  for (j=0; j<G_HEIGHT; j++)
	    {
	      if (ai_count_terr(player, i, j) == maxterr) N++;
	      if (N == nc)
		{
		  mvx = i;
		  mvy = j;
		  N++;
		}
	    }
	}
    }
  else
    {
      if (nshared == 1) nc = 1;
      else nc = (rand() % (nshared))+1;
      N = 0;
      for (i=0; i<G_WIDTH; i++)
	{
	  for (j=0; j<G_HEIGHT; j++)
	    {
	      if ((g_grid[i][j] & 3) == GP_SHARED) /* &3 to eliminate GT_STONE */
		{
		  N++;
		  if (N == nc)
		    {
		      mvx = i;
		      mvy = j;
		    }
		}
	    }
	}
    }
  
  *x = mvx;
  *y = mvy;
}

#define N_HARD_OPENINGS 9
static int hard_opening_x[N_HARD_OPENINGS] = { 2, 2, 21, 21,11,11,11, 2, 21 };
static int hard_opening_y[N_HARD_OPENINGS] = { 2,13,  2, 13, 2,13, 7, 7,  7 };

/* NORMAL/HARD STRATEGY
 *
 *  0. (Hard only) If any of the `opening' moves are unoccupied, pick
 *     one at random.
 *  1. Make the move that would gain the most territory.
 *     (i.e., stay away from the edges, and from opponent's territory slightly.)
 *  2. (Hard only) As above, but make certain that the opponent can't retaliate.
 *  3. If there are any shared squares, move to the most advantageous one.
 *
 *
 */
void ai_think_move(int player, int *x, int *y, int difficulty)
{
  int nmoves, ncount, nc, nr; /* nc=N current, nr=N random */
  int strategy = 4;
  int i,j,tmp;
  char buf[80];
  int xcc[384], ycc[384];

  SDL_PumpEvents(); /* Allow the mouse to move during long times of thought. */

  /* Pick the best move */
 strat1:
  nmoves = -1; ncount = 0;
  if (difficulty)
    {
      for (i=0; i<G_WIDTH; i++)
	{
	  for (j=0; j<G_HEIGHT; j++)
	    {
	      nc = ai_count_gain(player, i, j);
	      if (nc == nmoves) ncount++;
	      if (nc > nmoves)
		{
		  nmoves = nc; ncount = 1;
		}
	    }
	}
      if ((ncount <= 0) || (nmoves <= 0)) /* Can't attack */
	{
	  goto strat2;
	}
      nmoves = -450; ncount = 0;
    }
  for (i=0; i<G_WIDTH; i++)
    {
      for (j=0; j<G_HEIGHT; j++)
	{
	  if ((g_grid[i][j] & 3) == GP_SHARED)
	    {
	      if (difficulty)
		nc = lookahead_count_gain(player, i, j, difficulty);
	      else
		nc = ai_count_gain(player, i, j);
	      if (nc == nmoves)
		{
		  xcc[ncount] = i;
		  ycc[ncount] = j;
		  ncount++;
		}
	      if (nc > nmoves)
		{
		  xcc[0] = i;
		  ycc[0] = j;
		  nmoves = nc; ncount = 1;
		}
	    }
	}
    }
  if ((ncount <= 0)) /* Can't attack */
    {
      goto strat2; /* Considered harmful (: */
    }

  if (ncount == 1) nr = 0;
  else nr = (rand() % (ncount));
  nc=0;

  *x = xcc[nr];
  *y = ycc[nr];
  return;

  /* Grab the most non-enemy territory - NON-ATTACK MODE */
 strat2:
  if (!difficulty) goto strat3;
  if (G_WIDTH != 24) goto strat2b;

  /* Check opening moves. */
  nmoves = -1; ncount = 0;
  for (i=0; i<N_HARD_OPENINGS; i++)
    {
      if ((g_can_move(player, hard_opening_x[i], hard_opening_y[i])) &&
	  (!ai_would_conflict(player, hard_opening_x[i], hard_opening_y[i])))
	{
	  *x = hard_opening_x[i];
	  *y = hard_opening_y[i];
	  return;
	}
    }
  

 strat2b:
  nmoves = -450; ncount = 0;
  for (i=0; i<G_WIDTH; i++)
    {
      for (j=0; j<G_HEIGHT; j++)
	{
#if 0
	  if (ai_would_conflict(player, i, j)) break;
	  nc = ai_count_terr(player, i, j);
#endif
	  nc = lookahead_count_gain(player, i, j, difficulty);
	  if (nc == nmoves)
	    {
	      xcc[ncount] = i;
	      ycc[ncount] = j;
	      ncount++;
	    }
	  if (nc > nmoves)
	    {
	      xcc[0] = i;
	      ycc[0] = j;
	      nmoves = nc; ncount = 1;
	    }
	}
    }
  if ((ncount <= 0)) /* A problem */
    {
      goto strat3;
    }

  if (ncount == 1) nr = 0;
  else nr = (rand() % (ncount));
  nc=0;

  *x = xcc[nr];
  *y = ycc[nr];
  return;

  /* Grab the most territory */
 strat3:
  nmoves = -1; ncount = 0;
  for (i=0; i<G_WIDTH; i++)
    {
      for (j=0; j<G_HEIGHT; j++)
	{
	  nc = ai_count_terr(player, i, j);
	  if (nc == nmoves) ncount ++;
	  if (nc > nmoves)
	    {
	      nmoves = nc; ncount = 1;
	    }
	}
    }
  if (nmoves < 0) /* Can't happen */
    {
      return;
    }

  if (ncount == 1) nr = 1;
  else nr = (rand() % (ncount))+1;
  nc=0;

  for (i=0; i<G_WIDTH; i++)
    {
      for (j=0; j<G_HEIGHT; j++)
	{
	  tmp = ai_count_terr(player, i, j);
	  if (tmp == nmoves) nc++;
	  if (nc == nr)
	    {
	      *x = i;
	      *y = j;
	      return;
	    }
	}
    }

  return; /* Can't  happen */
}

void ai_move(int player, int level)
{
  int mvx=-1, mvy=-1;

  if (level == R_AI_EASY)
    {
      ai_easy_move(player, &mvx, &mvy);
    }
  else if (level == R_AI_NORMAL)
    {
      ai_think_move(player, &mvx, &mvy, 0);
    }
  else if (level == R_AI_HARD)
    {
      ai_think_move(player, &mvx, &mvy, AI_NASTINESS);
    }
  else
    {
      SDLhelp_Problem(bl_screen, "AI data has been frotzed, tell Cameron!");
      return;
    }

  if ((mvx < 0)||(mvy<0)||(mvx>=G_WIDTH)||(mvy>=G_HEIGHT))
    {
      SDLhelp_Problem(bl_screen, "AI bug, tell Cameron!");
      mvx = mvy = 0;
    }
  if (!g_can_move(player, mvx, mvy))
    {
      SDLhelp_Problem(bl_screen, "Don't look now, but the computer is cheating!");
    }
  
  r_flash_move(mvx,mvy);
}
