/* sdlhelp.c - Portability and other helpful routines for SDL. */

#include <SDL.h>
#include <string.h>
#ifdef WIN32
# include <windows.h>
#endif
#include "sdlhelp.h"

extern char *SDLhelp_fontdata[96];

/* Self-explanatory */
void SDLhelp_Error(char *m)
{
#ifdef WIN32
  MessageBox(NULL, m, "Error", MB_OK | MB_ICONEXCLAMATION);
#else
  fprintf(stderr, "Error: %s\n", m);
#endif
  exit(1);
}

/* Reports it on screen, waits for keypress, restores screen (*16-bit only*) */
void SDLhelp_Problem(SDL_Surface *scr, char *t)
{
  SDL_Event ev;
  Uint32 bg, fg;
  SDL_Surface *screenbuf;


  bg = SDL_MapRGB(scr->format, 0x20, 0x20, 0x20);
  fg = SDL_MapRGB(scr->format, 0x00, 0xA0, 0xD0);
  screenbuf = SDL_CreateRGBSurface(SDL_SWSURFACE, scr->w, scr->h, 16,
				   scr->format->Rmask,
				   scr->format->Gmask,
				   scr->format->Gmask,
				   scr->format->Amask);
  SDL_BlitSurface(scr, NULL, screenbuf, NULL);
  
  SDL_FillRect(scr, NULL, bg);
  SDLhelp_DrawString(scr, (scr->w/2)-6*4, (scr->h/10), "ERROR", fg);
  SDLhelp_DrawString(scr, (scr->w/2)-(strlen(t)*4), (scr->h/2), t, fg);
  SDLhelp_DrawString(scr, (scr->w/2)-9*4, scr->h-(scr->h/10), "PRESS KEY", fg);
  SDL_UpdateRect(scr, 0, 0, 0, 0);

  for (;;)
    {
      SDL_WaitEvent(&ev);
      if (ev.type == SDL_KEYUP) break;
    }

  SDL_BlitSurface(screenbuf, NULL, scr, NULL);
  SDL_FreeSurface(screenbuf);
  SDL_UpdateRect(scr, 0, 0, 0, 0);

}

/* Returns the directory that the executable file is
 * contained in. This can be used to locate data files.
 * Also, this is _not_ foolproof: if the programme is
 * not started in the same directory as its executable,
 * and the parent process doesn't set argv[0] properly,
 * this code will break.
 * The returned directory may be relative or absolute.
 * E.g., ".", "..", "/home/cp/projects/blah", "/usr/local/foo"
 * WARNING: argv0 will be modified!
 */
char *SDLhelp_LocateExecutable(char *argv0)
{
#ifdef WIN32
  return ".";
#else
  char *x;
  /* GLibc declares strrchr as:
       char *strrchr(char *haystack, char needle);
     (!!)
  */
  if ((x = strrchr(argv0, '/')) == NULL)
    return ".";
  *x = '\0';
  return argv0;
#endif
}

/* In Win32, the mouse can't be moved in SDL_Delay(). This fixes it. */
void SDLhelp_Delay(int n)
{
#ifndef WIN32
  SDL_Delay(n);
#else
  int st, ct;

  st = SDL_GetTicks() + n;
  while (SDL_GetTicks() < st) SDL_PumpEvents();
#endif
}

/* For some reason, SDL_StartTicks() causes problems in Win32, where it isn't
 * needed anyway.
 */
void SDLhelp_InitTicks()
{
#ifndef WIN32
  SDL_StartTicks();
#endif
}

/* Important points to remember:
 *   "colour" has a "u" in it.
 *
 *   If S is a hardware surface, it _must_ be locked before
 *   calling SDLhelp_DrawChar. (The docs say not to call library
 *   functions when the surface is locked, but ...)
 *
 *   If the bitmap isn't 16 bpp.... bad things may happen.
 *   Also, no checks are made to see if the character is
 *   being kept on-screen.  The programme will crash, possibly
 *   taking Windows down with it (assuming that you aren't running Linux).
 */
void SDLhelp_DrawChar(SDL_Surface *s, int x, int y, char ch, Uint32 colour)
{
  int i, j;
  char *data;
  int bitmask, bit;
  Uint16 *mempl, *memp;

  if (ch<32) ch=32;
  data=SDLhelp_fontdata[ch-32];

  mempl = ((Uint16 *)s->pixels) + ((y*s->pitch)>>1) + x;

  for (i=0; i<16; i++) /* Outer loop: draw lines */
    {
      bitmask = 0x80;
      memp = mempl;
      for (j=0; j<8; j++)
	{
	  bit = data[i] & bitmask;
	  if (bit != 0)
	    *memp = (Uint16) colour;
	  bitmask >>= 1;
	  memp++;
	}
      mempl += (s->pitch>>1);
    }

}

/* Simple wrapper around SDLhelp_DrawChar for accepting whole strings. */
void SDLhelp_DrawString(SDL_Surface *s, int x, int y, char *m, Uint32 colour)
{
  for(; *m != 0; m++, x+=8) SDLhelp_DrawChar(s, x, y, *m, colour);
}

/* Dissolve screen - works properly in a 16-bit mode. */
void SDLhelp_DissolveScreen(SDL_Surface *scr, int x, int y, Uint32 black)
{
  int i, j;
  Uint16 *fbuf;

  fbuf = (Uint16 *)scr->pixels;
  SDL_ShowCursor(0);

  for (i=0; i<y; i++)
    {
      for (j=(((i%2)==0)?0:1); j<x; j+=2)
	{
	  fbuf[((i*scr->pitch)>>1)+j] = black;
	}
      SDL_UpdateRect(scr, 0, i, x, 1);
    }
  for (i=y-1; i>=0; i--)
    {
      for (j=(((i%2)==0)?1:0); j<x; j+=2)
	{
	  fbuf[((i*scr->pitch)>>1)+j] = black;
	}
      SDL_UpdateRect(scr, 0, i, x, 1);
    }

  SDL_ShowCursor(1);
}

/* Cross fade - also 16-bit only */
void SDLhelp_CrossFade(SDL_Surface *scr, SDL_Surface *new, int x, int y)
{
  int i, j;
  Uint16 *fbuf, *nbuf;

  fbuf = (Uint16 *)scr->pixels;
  nbuf = (Uint16 *)new->pixels;

  SDL_ShowCursor(0);

  for (i=0; i<y; i++)
    {
      for (j=(((i%2)==0)?0:1); j<x; j+=2)
	{
	  fbuf[((i*scr->pitch)>>1)+j] = nbuf[((i*new->pitch)>>1)+j];
	}
      SDL_UpdateRect(scr, 0, i, x, 1);
    }
  for (i=y-1; i>=0; i--)
    {
      for (j=(((i%2)==0)?1:0); j<x; j+=2)
	{
	  fbuf[((i*scr->pitch)>>1)+j] = nbuf[((i*new->pitch)>>1)+j];
	}
      SDL_UpdateRect(scr, 0, i, x, 1);
    }

  SDL_ShowCursor(1);
}
